import numpy as np
import scipy.stats as stats
import os
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
from venn import venn

from sotodlib import hwp
import sotodlib.core as core
from sotodlib.core.flagman import has_all_cut, has_any_cuts, count_cuts, sparse_to_ranges_matrix

def plot_det_bias_flags(aman, det_bias_flags, rfrac_range=(0.1, 0.7),
                        psat_range=(0, 15), filename="./bias_cuts_venn.png"):
    """
    Function for plotting bias cuts.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    det_bias_flags : AxisManager
        Output msk_aman of tod_ops.flags.get_det_bias_flags with full_output=True.
    rfrac_range : Tuple
        Tuple (lower_bound, upper_bound) for rfrac det selection.
    psat_range : Tuple
        Tuple (lower_bound, upper_bound) for P_SAT from IV analysis.
        P_SAT in the IV analysis is the bias power at 90% Rn in pW.
    filename : str
        Full filename with direct path to plot output directory.
    """
    msk_names = ['bg', 'r_tes', 'r_frac_gt', 'r_frac_lt', 'p_sat_gt', 'p_sat_lt']
    msk = det_bias_flags['det_bias_flags']
    msks = []
    for name in msk_names:
        msks.append(det_bias_flags[f'{name}_flags'])
    all_bad_dets = has_all_cut(msk)
    msk_ids = []
    for msk in msks[:2]:
        bad_dets = has_all_cut(msk)
        msk_ids.append(np.where(bad_dets == True)[0])

    for i in range(0, 3, 2):
        bad_dets1 = has_all_cut(msks[i+2])
        bad_dets2 = has_all_cut(msks[i+3])
        booleans = [bad_dets1, bad_dets2]
        union = np.asarray(list(map(any, zip(*booleans))))
        msk_ids.append(np.where(union == True)[0])

    msk_dict = {'bg < 0': set(msk_ids[0]),
                'r_tes <= 0': set(msk_ids[1]),
                f'r_frac < {rfrac_range[0]} or > {rfrac_range[1]}': set(msk_ids[2]),
                f'p_sat*1e12 < {psat_range[0]} or > {psat_range[1]}': set(msk_ids[3])}

    venn(msk_dict)

    plt.title(f"{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min\nDetectors Cut per Range (Total cut: {len(np.where(all_bad_dets == True)[0])}/{len(aman.dets.vals)})")
    plt.tight_layout()
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)

def plot_4f_2f_counts(aman, modes=np.arange(1,49), filename='./4f_2f_counts.png'):
    """
    Function for plotting 4f/2f counts for each bandpass.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    modes : list of int
        The HWPSS harmonic modes to extract.
    filename : str
        Full filename with direct path to plot output directory.
    """
    hwpss_ratsatp1 = {}
    fig, axs = plt.subplots(3, 2, figsize=(15, 15))
    for i, band in enumerate(['f090', 'f150']):
        hwpss_ratsatp1[band] = {}
        m90s = (aman.det_info.wafer.bandpass == 'f090')
        m150s = (aman.det_info.wafer.bandpass == 'f150')
        m = [m90s, m150s]
        hwpss_aman = aman.restrict('dets', aman.dets.vals[m[i]], in_place=False)
        if hwpss_aman.dets.count == 0:
            print(f"No dets in {band} band")
            continue
        stats = hwp.hwp.get_hwpss(hwpss_aman, modes=modes, merge_stats=False, merge_model=False)
        a_4f = np.sqrt(stats.coeffs[:,6]**2 + stats.coeffs[:,7]**2)
        a_2f = np.sqrt(stats.coeffs[:,2]**2 + stats.coeffs[:,3]**2)
        hwpss_ratsatp1[band] = a_4f/a_2f

        hist, bins = np.histogram(hwpss_ratsatp1[band], bins=50)
        logbins = np.logspace(np.log10(bins[0]),np.log10(bins[-1]),len(bins))
        _ = axs[0, i].hist(hwpss_ratsatp1[band], bins = logbins,
                           color='C0', alpha=0.75)
        axs[0, i].axvline(np.median(hwpss_ratsatp1[band]), color = 'C0', ls = ':',
                          label = f"median: {np.median(hwpss_ratsatp1[band]):.2e}")
        axs[0, i].legend()
        axs[0, i].set_xscale('log')
        axs[0, i].set_title(f'Band {band}')
        axs[0, i].set_xlabel('A_4f/A_2f')
        
        hist, bins = np.histogram(a_4f, bins=50)
        logbins = np.logspace(np.log10(bins[0]),np.log10(bins[-1]),len(bins))
        axs[1, i].hist(a_4f, bins = logbins,
                       color='C0', alpha=0.75)
        axs[1, i].axvline(np.median(a_4f), color = 'C0', ls = ':',
                          label = f"median: {np.median(a_4f):.2e}")
        axs[1, i].legend()
        axs[1, i].set_xscale('log')
        axs[1, i].set_xlabel('A_4f')
        
        hist, bins = np.histogram(a_2f, bins=50)
        logbins = np.logspace(np.log10(bins[0]),np.log10(bins[-1]),len(bins))
        axs[2, i].hist(a_2f, bins = logbins,
                       color='C0', alpha=0.75)
        axs[2, i].axvline(np.median(a_2f), color = 'C0', ls = ':',
                          label = f"median: {np.median(a_2f):.2e}")
        axs[2, i].legend()
        axs[2, i].set_xscale('log')
        axs[2, i].set_xlabel('A_2f')
        
    for i in range(3):
        axs[i, 0].set_ylabel('Counts')

    plt.suptitle(f'{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min\n4f/2f Counts')
    plt.tight_layout()
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)

def plot_hwpss_fit_status(aman, hwpss_stats, plot_dets=None, plot_num_dets=3,
                          filename='./hwpss_stats.png'):
    """
    Function for plotting HWPSS fit status.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    hwpss_stats : AxisManager.hwpss_stats
        The HWPSS stats output.
    plot_dets : list
        List of dets to plot
    plot_num_dets : list
        Number of dets to plot.
    filename : str
        Full filename with direct path to plot output directory.
    """
    fig, ax = plt.subplots(1, 2, figsize=(15, 5))
    
    if plot_dets is None:
        plot_step = hwpss_stats.dets.count/(plot_num_dets)
        plot_dets_idx = np.arange(0, hwpss_stats.dets.count, plot_step).astype(int)
        plot_dets = hwpss_stats.dets.vals[plot_dets_idx]
    else:
        plot_dets_idx = np.where(np.isin(hwpss_stats.dets.vals, plot_dets))[0]

    for i, det_idx in enumerate(plot_dets_idx):
        ax[0].errorbar(hwpss_stats.binned_angle, hwpss_stats.binned_signal[det_idx], yerr=hwpss_stats.sigma_bin[det_idx],
                       alpha=0.5, color='tab:blue', fmt='.-', capsize=2, zorder=2, label='binned signal' if i ==0 else None)
        ax[0].plot(aman.hwp_angle[:2000], aman.signal[det_idx][:2000]-np.median(aman.signal[det_idx][:2000]),
                   alpha=0.5, color='tab:red', marker='o', markersize=0.5, linestyle='None', zorder=1,
                   label='unbinned signal (2000 samps)' if i ==0 else None)

        modes = [int(mode_name[1:]) for mode_name in list(hwpss_stats.modes.vals[::2])]
        ax[0].plot(hwpss_stats.binned_angle, hwpss_stats.binned_model[det_idx], 
                   alpha=0.9, color='tab:orange', zorder=3, label=f'binned model \n(modes = {modes})' if i ==0 else None)

    ax[0].legend()
    ax[0].set_xlabel('HWP angle [rad]')
    ax[0].set_ylabel('Signal [Readout Radians]')
    ax[0].set_title(f'random {plot_num_dets} detectors')

    ax[1].hist(hwpss_stats.redchi2s, bins=np.logspace(start=-1, stop=2, num=50))
    ax[1].axvline(x=np.nanmedian(hwpss_stats.redchi2s), linestyle='dashed', color='black',
                  label=f'median: {np.nanmedian(hwpss_stats.redchi2s):.2f}')
    ax[1].set_xscale('log')
    ax[1].set_yscale('log')
    ax[1].set_title(f'reduced chi2s distribution (Ndets={hwpss_stats.dets.count})')
    ax[1].legend()

    plt.suptitle(f'{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min\nHWPSS Stats\n')
    plt.subplots_adjust(top=0.85, bottom=0.2)
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)

def plot_sso_footprint(aman, planet_aman, sso, wafer_offsets=None, focal_plane=None, filename='./sso_footprint.png'):
    """
    Function for plotting SSO footprint.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    planet_aman : AxisManager
        Axis manager of results from obs_ops.sources.get_sso for a single planet.
    sso : str
        Name of planet.
    wafer_offsets : dict
        Dictionary of wafer offsets.
        Ex: wafer_offsets = {'ws0': (-2.5, -0.5),
                             'ws1': (-2.5, -13),
                             'ws2': (-13, -7),
                             'ws3': (-13, 5),
                             'ws4': (-2.5, 11.5),
                             'ws5': (8.5, 5),
                             'ws6': (8.5, -7)}
    focal_plane : str
        Path to focal plane file.
    filename : str
        Full filename with direct path to plot output directory.
    """
    xi_p = planet_aman['xi_p']
    eta_p = planet_aman['eta_p']

    if wafer_offsets is None:
        # Default wafer offsets
        wafer_offsets = {'ws0': (-2.5, -0.5),
                         'ws1': (-2.5, -13),
                         'ws2': (-13, -7),
                         'ws3': (-13, 5),
                         'ws4': (-2.5, 11.5),
                         'ws5': (8.5, 5),
                         'ws6': (8.5, -7)}

    if focal_plane is None:
        print('No focal plane file given.')
        return False
    else:
        hw = np.load(focal_plane)
    xi_hw, eta_hw, dets_hw = hw['xi_hw'], hw['eta_hw'], hw['dets_hw']
    
    fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8))

    ax.scatter(np.degrees(xi_hw), np.degrees(eta_hw), c='gray')
    ax.set_xlabel('$\\xi$ [degrees]')
    ax.set_ylabel('$\\eta$ [degrees]')
    ax.plot(np.degrees(xi_p), np.degrees(eta_p), color='C0', alpha=0.8)
    for k in wafer_offsets.keys():
        ax.text(wafer_offsets[k][0], wafer_offsets[k][1], k, fontsize=16)

    plt.suptitle(f'{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min\n{sso}')
    plt.tight_layout()
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)

def plot_pcabounds(aman, pca_aman, filename='./pca.png', signal=None, band=None, plot_ds_factor=20):
    """Subplot of pca bounds as well as the good and bad detector
    timestreams with 0th mode weight overplotted

    Parameters
    ----------
    aman : AxisManager
        input AxisManager
    pca_aman : AxisManager
        Relcal output AxisManager
    filename : str
        Full filename with direct path to plot output directory.
    signal : str or ndarray
        Signal name or signal array in aman. ``aman.signal`` is used if not provided.
    band : str
        Bandpass name. Assumes no bandpass separation if not provided.
    plot_ds_factor : int
        Factor to downsample signal plots. Default is 20.
    
    """
    if signal is None:
        signal = aman['signal']
    elif isinstance(signal, str):
        signal = aman[signal]
    elif isinstance(signal, np.ndarray):
        pass
    else:
        raise TypeError("Signal must be None, str, or ndarray")

    if band is None:
        xbounds = pca_aman.xbounds
        ybounds = pca_aman.ybounds
        modes = pca_aman.pca_mode0
    else:
        xbounds = pca_aman[f'{band}_xbounds']
        ybounds = pca_aman[f'{band}_ybounds']
        modes = pca_aman[f'{band}_pca_mode0']

    pca_dets = pca_aman.pca_det_mask
    good_indices = np.where(~pca_dets)[0]
    bad_indices = np.where(pca_dets)[0]

    fig = plt.figure(figsize=(10, 6))

    # Define axes
    ax1 = plt.subplot2grid((2, 2), (1, 0), colspan=1, rowspan=1)
    ax2 = plt.subplot2grid((2, 2), (1, 1), colspan=1, rowspan=1)
    ax3 = plt.subplot2grid((2, 2), (0, 0), colspan=2, rowspan=1)
    
    # ax1: good signals
    ax1.plot(aman.timestamps, modes, color='black', linewidth=3,
             label='0th mode', zorder=2, alpha=0.4)

    ax1.plot(aman.timestamps[::plot_ds_factor], np.divide(aman.signal[good_indices][:,::plot_ds_factor].T, pca_aman.pca_weight0[good_indices]), zorder=1, color='#D8BFD8', alpha=0.3)

    ax1.set_title(f'Good Detector Batch: ({len(good_indices)} dets)')
    ax1.legend(loc='upper left')
    ax1.grid()

    # ax2: bad signals
    ax2.plot(aman.timestamps, modes, color='black', linewidth=3,
             label='0th mode', zorder=2, alpha=0.4)

    ax2.plot(aman.timestamps[::plot_ds_factor], np.divide(aman.signal[bad_indices][:,::plot_ds_factor].T, pca_aman.pca_weight0[bad_indices]), zorder=1, color='#FFA07A', alpha=0.3)
        
    ax2.set_title(f'Bad Detector Batch: ({len(bad_indices)} dets)')
    ax2.legend(loc='upper left')
    ax2.grid()

    # ax3: box
    weight = np.abs(pca_aman.pca_weight0)
    ax3.plot(aman.det_cal.s_i[good_indices], weight[good_indices], '.', color='#D8BFD8', markersize=10,
             label=f'Good dets ({len(good_indices)} dets)', alpha=0.3)

    ax3.plot(aman.det_cal.s_i[bad_indices], weight[bad_indices], '.', color='#FFA07A', markersize=10,
             label=f'Bad dets ({len(bad_indices)} dets)', alpha=0.3)

    vertices = [(xbounds[0], ybounds[0]), (xbounds[1], ybounds[0]), (xbounds[1], ybounds[1]), (xbounds[0], ybounds[1])]
    box = matplotlib.patches.Polygon(vertices, closed=True, edgecolor='navy', linestyle='-.', fill=False, alpha=1, label='Boundary')
    ax3.add_patch(box)

    ax3.set_xlabel('Si')
    ax3.set_ylabel('0th Mode Weights')

    ax3.legend()
    ax3.grid()

    plt.suptitle(f'{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min\n{band}')
    plt.tight_layout()
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)

def plot_trending_flags(aman, trend_aman, filename='./trending_flags.png'):
    """
    Function for plotting trending flags.

    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    trend_aman : AxisManager
        Output trend_aman of tod_ops.flags.get_trending_flags with full_output=True.
    filename : str
        Full filename with direct path to plot output directory.
    """
    fig, axs = plt.subplots(1, 2, figsize=(5*2, 5*1))

    tdets = has_any_cuts(trend_aman.trend_flags)
    
    keep = np.abs(trend_aman.trends[~tdets, :]).T
    for i in range(keep.shape[1]):
        if i == 0:
            axs[0].plot(aman.timestamps[trend_aman.samp_start], keep[:, i], color = 'C0', 
                        alpha=0.75, marker='o', markersize=0.75, linestyle='None', label='Keep')
        else:
            axs[0].plot(aman.timestamps[trend_aman.samp_start], keep[:, i], color = 'C0', 
                        alpha=0.75, marker='o', markersize=0.75, linestyle='None')
    flagged = np.abs(trend_aman.trends[tdets, :]).T
    for i in range(flagged.shape[1]):
        if i == 0:
            axs[0].plot(aman.timestamps[trend_aman.samp_start], flagged[:, i], color = 'C1', 
                        marker='o', markersize=1.5, linestyle='None', label='Flagged')
        else:
            axs[0].plot(aman.timestamps[trend_aman.samp_start], flagged[:, i], color = 'C1', 
                        marker='o', markersize=1.5, linestyle='None')
    axs[0].set_yscale('log')
    axs[0].set_xlabel('Timestamp')
    axs[0].set_ylabel('Trend Slope')
    axs[0].set_title('Trending Slopes of Dets')
    axs[0].legend()

    axs[1].plot(aman.timestamps[::100], aman.signal[tdets][:,::100].T, color = 'C0', alpha = 0.5)
    axs[1].set_xlabel('Timestamp')
    axs[1].set_ylabel('Signal [Readout Radians]')
    axs[1].set_title(f'Trending Channels ({len(np.where(tdets == True)[0])} dets)')

    plt.suptitle(f"{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min\nTrending Flags (Total cut: {len(np.where(tdets == True)[0])}/{len(aman.dets.vals)})")
    plt.tight_layout()
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)

def plot_signal(aman, signal=None, xx=None, signal_name="signal", x_name="timestamps", plot_ds_factor=50, plot_ds_factor_dets=None, xlim=None, alpha=0.2, yscale='linear', y_unit=None, filename="./signal.png"):
    from operator import attrgetter
    if plot_ds_factor_dets is None:
        plot_ds_factor_dets = plot_ds_factor
    if signal is None:
        signal = attrgetter(signal_name)(aman)
    if xx is None:
        xx = attrgetter(x_name)(aman)
    yy = signal[::plot_ds_factor_dets, 1::plot_ds_factor].copy() # (dets, samps); (dets, nusamps); (dets, nusamps, subscans)
    xx = xx[1::plot_ds_factor].copy() # (samps); (nusamps)
    if x_name == "timestamps":
        xx -= xx[0]
    if yy.ndim > 2: # Flatten subscan axis into dets
        yy = yy.swapaxes(1,2).reshape(-1, yy.shape[1])

    if xlim is not None:
        xinds = np.logical_and(xx >= xlim[0], xx <= xlim[1])
        xx = xx[xinds]
        yy = yy[:,xinds]

    fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8))
    ax.plot(xx, yy.T, color='k', alpha=0.2)
    ax.set_yscale(yscale)
    if "freqs" in x_name:
        ax.set_xlabel("freq [Hz]")
    else:
        ax.set_xlabel(f"{x_name} [s]")
    y_unit = "" if y_unit is None else f" [{y_unit}]"
    ax.set_ylabel(f"{signal_name.replace('.Pxx', '')}{y_unit}")
    plt.suptitle(f"{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min")
    plt.tight_layout()
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)

def plot_psd(aman, signal=None, xx=None, signal_name="psd.Pxx", x_name="psd.freqs", plot_ds_factor=4, plot_ds_factor_dets=20, xlim=None, alpha=0.2, yscale='log', y_unit=None, filename="./psd.png"):
    return plot_signal(aman, signal, xx, signal_name, x_name, plot_ds_factor, plot_ds_factor_dets, xlim, alpha, yscale, y_unit, filename)

def plot_signal_diff(aman, flag_aman, flag_type="glitches", flag_threshold=10, plot_ds_factor=50, filename="./glitch_signal_diff.png"):
    """
    Function for plotting the difference in signal before and after cuts from either glitches or jumps.
    
    Parameters
    ----------
    aman : AxisManager
        Input axis manager.
    flag_aman : AxisManager
        Output jump_aman of tod_ops.jumps.jumps_aman or
        glitch_aman of tod_ops.flags.get_glitch_flags with full_output=True.
    flag_type : str
        Flag type to plot. Options: ["glitches, "jumps"]. Default is "glitches"
    flag_threshold : int
        Threshold to cut dets. Equivalent to "max_n_glitch" and "max_n_jumps".
    plot_ds_factor : int
        Factor to downsample signal plots. Default is 50.
    filename : str
        Full filename with direct path to plot output directory.
    """
    if flag_type == "glitches":
        flags = flag_aman.glitch_flags
        plot_name = f"Glitch Flags Signal Diff (Every {plot_ds_factor} dets)"
    elif flag_type == "jumps":
        flags = flag_aman.jump_flag
        plot_name = f"Jump Flags Signal Diff (Every {plot_ds_factor} dets)"
    else:
        raise ValueError("Flag type not recognized. Must be 'glitches' or 'jumps'")
    
    fig, axes = plt.subplots(1, 2, figsize=(15, 5))
    ax = axes.flatten()
    
    ax[0].plot(aman.timestamps[::plot_ds_factor], aman.signal[:,::plot_ds_factor].T, color='tab:blue', alpha=0.3)
    ax[0].set_xlabel("Timestamp")
    ax[0].set_ylabel("Signal [Readout Radians]")
    ax[0].set_title("Before")
    
    if flag_type == "glitches":
        flags = sparse_to_ranges_matrix(flag_aman.glitch_detection > flag_threshold)
    n_cut = count_cuts(flags)
    keep = n_cut <= flag_threshold
    flags_removed = aman.restrict("dets", aman.dets.vals[keep], in_place=False)

    ax[1].plot(flags_removed.timestamps[::plot_ds_factor], flags_removed.signal[:,::plot_ds_factor].T, color='tab:blue', alpha=0.3)
    ax[1].set_xlabel("Timestamp")
    ax[1].set_ylabel("Signal [Readout Radians]")
    ax[1].set_title("After")
    
    plt.suptitle(f"{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min\n{plot_name} (Total cut: {len(np.where(~keep)[0])}/{len(aman.dets.vals)})")
    plt.subplots_adjust(top=0.70, bottom=0.15)
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)

def plot_flag_stats(aman, flag_aman, flag_type="glitches", N_bins=30, filename="./glitch_stats.png"):
    """
    Function for plotting the glitches or jumps flags/cut statistics using the built in stats functions
    in the RangesMatrices class.
    Args:
    -----
    aman : AxisManager
        Input axis manager.
    flag_aman : AxisManager
        Output jump_aman of tod_ops.jumps.jumps_aman or
        glitch_aman of tod_ops.flags.get_glitch_flags with full_output=True.
    flag_type : str
        Flag type to plot. Options: ["glitches, "jumps"]. Default is "glitches"
    N_bins (int): Number of bins in the histogram.
    filename : str
        Full filename with direct path to plot output directory.
    """
    if flag_type == "glitches":
        flags = flag_aman.glitch_flags
        plot_name = "Glitch Stats"
    elif flag_type == "jumps":
        flags = flag_aman.jump_flag
        plot_name = "Jumps Stats"
    else:
        raise ValueError("Flag type not recognized. Must be 'glitches' or 'jumps'")
    
    fig, axes = plt.subplots(1, 2, figsize=(15, 5))
    ax = axes.flatten()
    frac_samp_glitches = (
        100 * np.asarray(flags.get_stats()["samples"]) / aman.samps.count
    )
    glitchlog = np.log10(frac_samp_glitches[frac_samp_glitches > 0])
    if glitchlog.size == 0:
        raise ValueError("No flags found (empty array).")
    binmin = int(np.floor(np.min(glitchlog)))
    binmax = int(np.ceil(np.max(glitchlog)))
    _ = ax[0].hist(
        frac_samp_glitches, bins=np.logspace(binmin, binmax, N_bins), label="_nolegend_"
    )
    medsamps = np.median(frac_samp_glitches)
    ax[0].axvline(medsamps, color="C1", ls=":", lw=2, label=f"Median: {medsamps:.2e}%")
    meansamps = np.mean(frac_samp_glitches)
    ax[0].axvline(meansamps, color="C2", ls=":", lw=2, label=f"Mean: {meansamps:.2e}%")
    modesamps = stats.mode(frac_samp_glitches, keepdims=True)
    ax[0].axvline(
        modesamps[0][0],
        color="C3",
        ls=":",
        lw=2,
        label=f"Mode: {modesamps[0][0]:.2e}%, Counts: {modesamps[1][0]}",
    )
    stdsamps = np.std(frac_samp_glitches)
    ax[0].axvspan(
        meansamps - stdsamps,
        meansamps + stdsamps,
        color="wheat",
        alpha=0.2,
        label=f"$\sigma$: {stdsamps:.2e}%",
    )
    ax[0].legend()
    ax[0].set_xlim(10**binmin, 10**binmax)
    ax[0].set_xscale("log")
    ax[0].set_xlabel("Fraction of Samples Flagged\nper Detector [%]", fontsize=16)
    ax[0].set_ylabel("Counts", fontsize=16)
    ax[0].set_title(
        "Samples Flagged Stats\n$N_{\mathrm{dets}}$ = "
        + f"{aman.dets.count}"
        + " and $N_{\mathrm{samps}}$ = "
        + f"{aman.samps.count}"
    )

    interval_glitches = np.asarray(flags.get_stats()["intervals"])
    binlinmax = np.quantile(interval_glitches, 0.98)
    _ = ax[1].hist(np.clip(interval_glitches, 0, binlinmax), bins=np.linspace(0, binlinmax, N_bins))
    medints = np.median(interval_glitches)
    ax[1].axvline(
        medints, color="C1", ls=":", lw=2, label=f"Median: {medints:.2e} intervals"
    )
    meanints = np.mean(interval_glitches)
    ax[1].axvline(
        meanints, color="C2", ls=":", lw=2, label=f"Mean: {meanints:.2e} intervals"
    )
    modeints = stats.mode(interval_glitches, keepdims=True)
    ax[1].axvline(
        modeints[0][0],
        color="C3",
        ls=":",
        lw=2,
        label=f"Mode: {modeints[0][0]:.2e} intervals, Counts: {modeints[1][0]}",
    )
    stdints = np.std(interval_glitches)
    ax[1].axvspan(
        meanints - stdints,
        meanints + stdints,
        color="wheat",
        alpha=0.2,
        label=f"$\sigma$: {stdints:.2e} intervals",
    )

    ax[1].legend()
    ax[1].set_xlim(-1, binlinmax)
    ax[1].set_xlabel("Number of Flag Intervals\nper Detector", fontsize=16)
    ax[1].set_ylabel("Counts", fontsize=16)
    ax[1].set_title(
        "Ranges Flag Manager Stats\n$N_{\mathrm{dets}}$ with $\geq$ 1 interval = "
        + f"{len(interval_glitches[interval_glitches > 0])}/{aman.dets.count}\n(98th quantile bin max)"
    )
    
    plt.suptitle(f"{aman.obs_info.obs_id}, dT = {np.ptp(aman.timestamps)/60:.1f} min\n{plot_name}")
    plt.subplots_adjust(top=0.70, bottom=0.15)
    head_tail = os.path.split(filename)
    os.makedirs(head_tail[0], exist_ok=True)
    plt.savefig(filename)



import numpy as np
import so3g
from spt3g import core
from sotodlib import core
import logging

logger = logging.getLogger(__name__)


def I_to_P_param(bandpass="f090", loading=10, inc_ang=5):
    """
    - HWP I to P leakage simulation function
    - Mueller matirices of three layer achromatic HWP \
      for each frequency and each incident angle are derived by RCWA simulation.
    - AR coating based on Mullite-Duroid coating for SAT1 at UCSD.\
      Thickness are used fabricated values.
    - 2f and 4f amplitudes and phases are extracted by HWPSS fitting of MIQ and MIU.
    - Amplitudes and phases are fitted by polynominal functions w.r.t each incident angle.
    - Only support MF because UHF and LF are design/fabrication phase. \
      We will add them after the fabrications.
    
    Args
    -----
    bandpass: str
        an 3-digit (front zero padded) integer "fXYZ" 
        where XYZ = 030/040/090/150/220/280
    loading: float
        intensity [Kcmb]
    inc_ang: float
        incident angle [deg]

    Returns
    --------
    amp_2f: float
        2f amplitude [mKcmb]
    amp_4f: float
        4f amplitude [mKcmb]
    phi_2f: float
        2f phase shift [rad]
    phi_4f: float
        4f phase shift [rad]
    """
    if bandpass == "f090":
        poly_A2_ang = np.array([-7.01e-06, -6.87e-05, 3.435e-02])
        poly_A4_ang = np.array([1.61e-07, 1.20e-05, 3.42e-06, 1.23e-07])
        phi_2f, phi_4f = 66.46, 55.57  # [deg]
    elif band == "f150":
        poly_A2_ang = np.array([-1.11e-05, 2.58e-05, 3.40e-02])
        poly_A4_ang = np.array([4.38e-08, 8.90e-06, 1.85e-06, 1.18e-07])
        phi_2f, phi_4f = 71.62, 54.43  # [deg]
    else:
        logger.error("Currently only supporting MF(f090/f150), stay tuned.")
    amp_2f = loading * np.poly1d(poly_A2_ang)(inc_ang)
    amp_4f = loading * np.poly1d(poly_A4_ang)(inc_ang)
    amp_2f, amp_4f = amp_2f * 1e3, amp_4f * 1e3
    phi_2f, phi_4f = np.deg2rad(phi_2f), np.deg2rad(phi_4f)
    if inc_ang == 0:
        amp_4f, phi_4f = 0, 0
    return amp_2f, amp_4f, phi_2f, phi_4f


def sim_hwpss(aman, name='hwpss_sim', hwp_freq=2., 
              bandpass='090', loading=10., inc_ang=5.,
              amp_2f_r=0.05, amp_4f_r=0.05, 
              phi_2f_r=0.1, phi_4f_r=0.1):
    """
    - HWPSS simulation function using 2f and 4f amp and phase from I_to_P_param function.
    - The simulated HWSS is added to the input AxisManager as “name”.

    Args
    -----
    aman: AxisManager
        target AxisManager
    hwp_freq: float
        HWP rotation speed, 2 [Hz] (default)
    bandpass: str
        an 3-digit (front zero padded) integer "XYZ" 
        where XYZ = 030/040/090/150/220/280
    loading: float
        intensity [Kcmb]
    inc_ang: float 
        incident angle [deg]
    amp_2f_r: float, optional
        2f HWPSS amplitude fluctuation
    amp_4f_r: float, optional
        4f HWPSS amplitude fluctuation
    phi_2f_r: float, optional
        2f HWPSS phase shift fluctuation
    phi_4f_r: float, optional
        4f HWPSS phase shift fluctuation

    """
    if name in aman.keys():
        aman.move(name, '')
    aman.wrap_new(name, ('dets', 'samps'))
    hwp_angle = np.mod(2 * np.pi * aman.timestamps * hwp_freq, 2 * np.pi)
    if 'hwp_angle' not in aman.keys():
        aman.wrap('hwp_angle', hwp_angle, [(0, 'samps')])
    else:
        aman.hwp_angle = hwp_angle
    nf = [2, 4]
    amp_2f, amp_4f, phi_2f, phi_4f = I_to_P_param(inc_ang=15)
    amp = [amp_2f, amp_4f]
    amp_r = [amp_2f * amp_2f_r, amp_4f * amp_4f_r]
    phi = [phi_2f, phi_4f]
    phi_r = [phi_2f * phi_2f_r, phi_4f * phi_4f_r]
    for n in range(len(nf)):
        amp_rnd = np.random.normal(
            loc=amp[n], scale=amp_r[n], size=(
                aman.dets.count,))
        phase_rnd = np.mod(
            np.random.normal(
                loc=phi[n],
                scale=phi_r[n],
                size=(
                    aman.dets.count,
                )),
            2 * np.pi)
        aman[name] += amp_rnd[:, None] * \
            np.cos(nf[n] * aman.hwp_angle + phase_rnd[:, None])


def sim_hwpss_2f4f(aman, name='hwpss_sim', hwp_freq=2., bandpass="090",
                   amp_2f=300, amp_4f=30, phi_2f=0, phi_4f=0,
                   amp_2f_r=0.05, amp_4f_r=0.05, phi_2f_r=0.1, phi_4f_r=0.1):
    """
    - HWPSS simulation function using given 2f and 4f amp and phase.
    - The simulated HWSS is added to the input AxisManager as “name”.

    Args
    -----
    aman: AxisManager
        target AxisManager
    hwp_freq: float
        2 [Hz] (default)
    bandpass: str 
        an 3-digit (front zero padded) integer "fXYZ"
        where XYZ = 030/040/090/150/220/280
    amp_2f: float
        2f HWPSS amplitude [Kcmb]
    amp_4f: float
        4f HWPSS amplitude [Kcmb]
    phi_2f: float
        2f HWPSS phase shift [deg.]
    phi_4f: float
        4f HWPSS phase shift [deg.]
    amp_2f_r: float 
        2f HWPSS amplitude fluctuation
    amp_4f_r: float
        4f HWPSS amplitude fluctuation
    phi_2f_r: float
        2f HWPSS phase shift fluctuation
    phi_4f_r: float 
        4f HWPSS phase shift fluctuation
    """
    if name in aman.keys():
        aman.move(name, '')
    aman.wrap_new(name, ('dets', 'samps'))
    hwp_angle = np.mod(2 * np.pi * aman.timestamps * hwp_freq, 2 * np.pi)
    if 'hwp_angle' not in aman.keys():
        aman.wrap('hwp_angle', hwp_angle, [(0, 'samps')])
    else:
        aman.hwp_angle = hwp_angle
    nf = [2, 4]
    amp = [amp_2f, amp_4f]
    amp_r = [amp_2f * amp_2f_r, amp_4f * amp_4f_r]
    phi = [phi_2f, phi_4f]
    phi_r = [phi_2f * phi_2f_r, phi_4f * phi_4f_r]
    for n in range(len(nf)):
        amp_rnd = np.random.normal(
            loc=amp[n], scale=amp_r[n], size=(
                aman.dets.count,))
        phase_rnd = np.mod(
            np.random.normal(
                loc=phi[n],
                scale=phi_r[n],
                size=(
                    aman.dets.count,
                )),
            2 * np.pi)
        aman[name] += amp_rnd[:, None] * \
            np.cos(nf[n] * aman.hwp_angle + phase_rnd[:, None])



# Copyright (c) 2018-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import os
import pickle

import ephem
import h5py
import healpy as hp
import numpy as np
import traitlets
from astropy import constants
from astropy import units as u
from scipy.constants import au as AU
from scipy.interpolate import RectBivariateSpline
from toast import qarray as qa
from toast.data import Data
from toast.observation import default_values as defaults
from toast.ops.operator import Operator
from toast.timing import function_timer, Timer
from toast.traits import (Bool, Float, Instance, Int, Quantity, Unicode,
                          trait_docs)
from toast.utils import Environment, Logger, unit_conversion, name_UID

XAXIS, YAXIS, ZAXIS = np.eye(3)


@trait_docs
class SimReadout(Operator):
    """Simulate various readout-related systematics"""

    API = Int(0, help="Internal interface version for this operator")

    realization = Int(0, help="The noise realization index")

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for telescope flags to use",
    )

    shared_flag_mask_readout = Int(
        defaults.shared_mask_invalid, help="Bit mask value for optional shared flagging"
    )

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for simulated signal",
    )

    glitch_rate = Quantity(0.1 * u.Hz, help="Glitch occurrence rate")

    glitch_amplitude_center = Quantity(
        1 * u.K, help="Center of Gaussian distribution for glitch amplitude"
    )

    glitch_amplitude_sigma = Quantity(
        1 * u.K, help="Width of Gaussian distribution for glitch amplitude"
    )

    bias_line_glitch_rate = Quantity(0.1 * u.Hz, help="Glitch occurrence rate")

    bias_line_glitch_amplitude_center = Quantity(
        1 * u.K, help="Center of Gaussian distribution for glitch amplitude"
    )

    bias_line_glitch_amplitude_sigma = Quantity(
        1 * u.K, help="Width of Gaussian distribution for glitch amplitude"
    )

    bias_line_glitch_amplitude_scatter = Float(
        1.0,
        help="Width of Gaussian distribution for *relative* glitch amplitude.  "
        "Negative factors get cropped."
    )

    cosmic_ray_glitch_rate = Quantity(
        0.000001 * u.Hz / u.mm**2,
        help="Cosmic ray glitch rate *either* per unit focalplane area "
        "e.g. [Hz / mm^2] *or* rate per the entire focalplane [Hz]",
    )

    cosmic_ray_glitch_amplitude_center = Quantity(
        1 * u.K, help="Center of Gaussian distribution for glitch amplitude"
    )

    cosmic_ray_glitch_amplitude_sigma = Quantity(
        1 * u.K, help="Width of Gaussian distribution for glitch amplitude"
    )

    cosmic_ray_correlation_length = Quantity(
        10 * u.mm, help="Distance scale of cosmic ray events on the focalplane"
    )

    cosmic_ray_time_constant = Quantity(
        1.0 * u.s, help="Dissipation time of cosmic ray events on the focaplane"
    )

    simulate_glitches = Bool(True, help="Enable glitch simulation")

    jump_rate = Quantity(0.001 * u.Hz, help="Jump occurrence rate")

    jump_amplitude_center = Quantity(
        0 * u.mK, help="Center of Gaussian distribution for jump amplitude"
    )

    jump_amplitude_sigma = Quantity(
        1 * u.mK, help="Width of Gaussian distribution for jump amplitude"
    )

    bias_line_jump_rate = Quantity(0.001 * u.Hz, help="Jump occurrence rate")

    bias_line_jump_amplitude_center = Quantity(
        0 * u.mK, help="Center of Gaussian distribution for jump amplitude"
    )

    bias_line_jump_amplitude_sigma = Quantity(
        1 * u.mK, help="Width of Gaussian distribution for jump amplitude"
    )

    bias_line_jump_amplitude_scatter = Float(
        1.0,
        help="Width of Gaussian distribution for *relative* jump amplitude.  "
        "Negative factors get cropped."
    )

    simulate_jumps = Bool(True, help="Enable jump simulation")

    misidentify_bolometers = Bool(True, help="Enable bolometer misidentification")

    misidentification_width = Quantity(
        1.0 * u.deg,
        help="Probability of misidentifying two channels is a Gaussian "
        "function of their distance on the focalplane.",
    )

    yield_center = Float(
        0.8, help="Center of Gaussian distribution for detector yield"
    )

    yield_sigma = Float(
        0.1, help="Width of Gaussian distribution for detector yield"
    )

    simulate_yield = Bool(True, help="Simulate failed bolometers")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _simulate_events(self, seed, nsample, events_per_sample):
        """Simulate a Poissonian distribution of events"""
        np.random.seed(int(seed) % 2**32)
        event_counts = np.random.poisson(events_per_sample, nsample)
        event_indices = np.argwhere(event_counts).ravel()
        event_counts = event_counts[event_indices].copy()
        return event_counts, event_indices

    def _simulate_amplitudes(
            self, seed, event_counts, amplitude_center, amplitude_sigma
    ):
        """Simulate Gaussian event amplitudes"""
        # Drawing amplitudes for every affected sample is not the
        # same as drawing amplitudes for every event but it is fast.
        # If we routinely have more than one event occurring during
        # the same time stamp we are in trouble
        np.random.seed(int(seed) % 2**32)
        n = event_counts.size
        amplitudes = amplitude_center + np.random.randn(n) * amplitude_sigma
        events = amplitudes * event_counts
        return events

    def _add_uncorrelated_glitches(
            self, comm, signal, obs_id, focalplane, fsample, nsample, local_dets
    ):
        """Simulate uncorrelated electronic glitches"""
        log = Logger.get()

        events_per_sample = self.glitch_rate.to_value(u.Hz) / fsample.to_value(u.Hz)
        nglitch = 0
        for det in local_dets:
            det_id = focalplane[det]["uid"]
            sig = signal[det]
            event_counts, event_indices = self._simulate_events(
                obs_id + det_id  + self.realization + 8123765,
                nsample,
                events_per_sample,
            )
            events = self._simulate_amplitudes(
                obs_id + det_id  + self.realization + 150243972,
                event_counts,
                self.glitch_amplitude_center.to_value(signal.units),
                self.glitch_amplitude_sigma.to_value(signal.units),
            )
            # Add delta-function glitches
            sig[event_indices] += events
            nglitch += np.sum(event_counts)

        if comm is not None:
            nglitch = comm.allreduce(nglitch)
        log.debug_rank(f"Simulated {nglitch} uncorrelated glitches", comm=comm)

        return

    def _add_correlated_glitches(
            self, comm, signal, obs_id, focalplane, fsample, nsample, bias2det
    ):
        """Simulate bias line glitches"""
        log = Logger.get()

        events_per_sample = self.bias_line_glitch_rate.to_value(u.Hz) \
                            / fsample.to_value(u.Hz)
        nglitch = 0
        for tube, wafers in bias2det.items():
            tube_id = name_UID(tube)
            for wafer, biases in wafers.items():
                wafer_id = name_UID(wafer)
                for bias, dets in biases.items():
                    # The event locations on the bias line are shared
                    event_counts, event_indices = self._simulate_events(
                        obs_id + tube_id + wafer_id + bias  + self.realization \
                        + 5295629,
                        nsample,
                        events_per_sample,
                    )
                    # Base amplitude is shared ...
                    events = self._simulate_amplitudes(
                        obs_id + tube_id + wafer_id  + self.realization + 150243972,
                        event_counts,
                        self.bias_line_glitch_amplitude_center.to_value(signal.units),
                        self.bias_line_glitch_amplitude_sigma.to_value(signal.units),
                    )
                    # ... but modulated at the detector level
                    for det in dets:
                        det_id = focalplane[det]["uid"]
                        sig = signal[det]
                        scale = self._simulate_amplitudes(
                            obs_id + det_id  + self.realization + 74712734,
                            event_counts,
                            1,
                            self.bias_line_glitch_amplitude_scatter,
                        )
                        scale[scale < 0] = 0
                        # Add delta-function glitches
                        sig[event_indices] += events * scale
                    nglitch += np.sum(event_counts)

        if comm is not None:
            nglitch = comm.allreduce(nglitch)
        log.debug_rank(f"Simulated {nglitch} correlated glitches", comm=comm)

        return

    def _add_cosmic_ray_glitches(
            self,
            comm,
            signal,
            obs_id,
            focalplane,
            fsample,
            nsample,
            local_dets,
            det2position,
    ):
        """Simulate cosmic ray glitches

        This is the only glitch population to be convolved with a time constant
        """
        log = Logger.get()

        platescale = focalplane.detector_data.meta["platescale"].to_value(u.deg / u.mm)
        fov = focalplane.field_of_view.to_value(u.deg)
        diameter = fov / platescale

        # Glitches are simulated over a square grid and individual
        # detectors are affected based on their proximity to the event
        area = diameter**2
        ratio = np.pi / 4  # square to disc area ratio
        try:
            # Did the user supply a rate for the entire focaplane?
            glitch_rate = self.cosmic_ray_glitch_rate.to_value(u.Hz)
            # Increase the rate to sample the glitches across a square area
            glitch_rate /= ratio
        except u.UnitConversionError:
            # Nope, it must be per unit area
            glitch_rate = self.cosmic_ray_glitch_rate.to_value(u.Hz / u.mm**2)
            glitch_rate *= area

        log.debug_rank(
            f"FOV = {fov:.2f} deg, platescale = {platescale:.6f} deg/mm, \n"
            f"focalplane diameter = {diameter} mm, area = {area * ratio:.3f} mm^2\n"
            f"glitch_rate = {glitch_rate / area:.6f} Hz/mm^2, \n"
            f"glitch_rate = {glitch_rate * ratio:.6f} Hz / focalplane",
            comm=comm,
        )

        events_per_sample = glitch_rate / fsample.to_value(u.Hz)
        event_counts, event_indices = self._simulate_events(
            obs_id + self.realization + 98552896,
            nsample,
            events_per_sample,
        )

        # Generate a glitch profile that we will re-use
        tau = self.cosmic_ray_time_constant
        wtemplate = int(10 * tau.to_value(u.s) * fsample.to_value(u.Hz))
        log.debug_rank(f"Width of the glitch template is {wtemplate}", comm=comm)
        t = np.arange(wtemplate) / fsample.to_value(u.Hz)
        template = np.exp(-t / tau.to_value(u.s))

        # For every event, we need a location
        sqradius = (0.5 * diameter)**2
        center = self.cosmic_ray_glitch_amplitude_center.to_value(signal.units)
        sigma = self.cosmic_ray_glitch_amplitude_sigma.to_value(signal.units)
        corr_length = self.cosmic_ray_correlation_length.to_value(u.mm)
        nglitch = 0
        ndet = 0
        for start, count in zip(event_indices, event_counts):
            stop = min(start + wtemplate, nsample)
            ind = slice(start, stop)
            templ = template[:stop - start]
            for i in range(count):
                amp = center + np.random.randn() * sigma
                x = (np.random.rand() - .5) * diameter
                y = (np.random.rand() - .5) * diameter
                if x**2 + y**2 > sqradius:
                    continue
                nglitch += 1
                # Observe the glitch with the local detectors
                for det in local_dets:
                    xdet, ydet = det2position[det]
                    dist = np.sqrt((x - xdet)**2 + (y - ydet)**2)
                    if dist > 4 * corr_length:
                        # This detector is to far to register
                        continue
                    ndet += 1
                    amp_det = amp * np.exp(-dist / corr_length)
                    signal[det][ind] += amp_det * templ

        if comm is not None:
            ndet = comm.allreduce(ndet)
        log.debug_rank(
            f"Simulated {nglitch} cosmic ray glitches and {ndet} "
            f"detector events",
            comm=comm,
        )

        return

    @function_timer
    def _add_glitches(
            self,
            comm,
            obs_id,
            times,
            focalplane,
            signal,
            all_dets,
            local_dets,
            bias2det,
            det2position,
    ):
        """Simulate glitches"""
        if not self.simulate_glitches:
            return

        fsample = focalplane.sample_rate
        nsample = times.size

        if self.glitch_rate != 0:
            self._add_uncorrelated_glitches(
                comm, signal, obs_id, focalplane, fsample, nsample, local_dets
            )

        if self.bias_line_glitch_rate != 0:
            self._add_correlated_glitches(
                comm, signal, obs_id, focalplane, fsample, nsample, bias2det
            )

        if self.cosmic_ray_glitch_rate != 0:
            self._add_cosmic_ray_glitches(
                comm,
                signal,
                obs_id,
                focalplane,
                fsample,
                nsample,
                local_dets,
                det2position,
            )

        return

    @function_timer
    def _add_jumps(
            self,
            comm,
            obs_id,
            times,
            focalplane,
            signal,
            all_dets,
            local_dets,
            bias2det,
    ):
        """Simulate baseline jumps"""
        if not self.simulate_jumps:
            return
        log = Logger.get()
        fsample = focalplane.sample_rate
        nsample = times.size

        # Uncorrelated jumps
        events_per_sample = self.jump_rate.to_value(u.Hz) / fsample.to_value(u.Hz)
        njump = 0
        for det in local_dets:
            det_id = focalplane[det]["uid"]
            sig = signal[det]
            event_counts, event_indices = self._simulate_events(
                obs_id + det_id  + self.realization + 45835364,
                nsample,
                events_per_sample,
            )
            events = self._simulate_amplitudes(
                obs_id + det_id  + self.realization + 250243972,
                event_counts,
                self.jump_amplitude_center.to_value(signal.units),
                self.jump_amplitude_sigma.to_value(signal.units),
            )
            # Change baseline at every event
            for index, event in zip(event_indices, events):
                sig[index:] += event
            njump += np.sum(event_counts)

        if comm is not None:
            njump = comm.allreduce(njump)
        log.debug_rank(f"Simulated {njump} uncorrelated jumps", comm=comm)

        # Bias line jumps
        events_per_sample = self.bias_line_jump_rate.to_value(u.Hz) \
                            / fsample.to_value(u.Hz)
        njump = 0
        for tube, wafers in bias2det.items():
            tube_id = name_UID(tube)
            for wafer, biases in wafers.items():
                wafer_id = name_UID(wafer)
                for bias, dets in biases.items():
                    # The event locations on the bias line are shared
                    event_counts, event_indices = self._simulate_events(
                        obs_id + tube_id + wafer_id + bias  + self.realization + 936582,
                        nsample,
                        events_per_sample,
                    )
                    # Base amplitude is shared ...
                    events = self._simulate_amplitudes(
                        obs_id + tube_id + wafer_id  + self.realization + 1295723,
                        event_counts,
                        self.bias_line_jump_amplitude_center.to_value(signal.units),
                        self.bias_line_jump_amplitude_sigma.to_value(signal.units),
                    )
                    # ... but modulated at the detector level
                    for det in dets:
                        det_id = focalplane[det]["uid"]
                        sig = signal[det]
                        scale = self._simulate_amplitudes(
                            obs_id + det_id  + self.realization + 83276345,
                            event_counts,
                            1,
                            self.bias_line_jump_amplitude_scatter,
                        )
                        scale[scale < 0] = 0
                        # Change baseline at every event
                        for index, event in zip(event_indices, events * scale):
                            sig[index:] += event
                    njump += np.sum(event_counts)

        if comm is not None:
            njump = comm.allreduce(njump)
        log.debug_rank(f"Simulated {njump} correlated jumps", comm=comm)

        return

    @function_timer
    def _fail_bolometers(
            self,
            comm,
            obs_id,
            times,
            focalplane,
            signal,
            all_dets,
            local_dets,
            bias2det,
    ):
        """Simulate bolometers that fail to bias or otherwise are unusable"""
        if not self.simulate_yield:
            return
        log = Logger.get()
        np.random.seed(int(obs_id + self.realization + 59127574) % 2**32)
        yield_ = self.yield_center + np.random.randn() * self.yield_sigma
        yield_ = min(1, max(0, yield_))
        nfail = 0
        for det in local_dets:
            det_id = focalplane[det]["uid"]
            np.random.seed(int(obs_id + self.realization + 59127574 + det_id) % 2**32)
            x = np.random.rand()
            if x > yield_:
                # FIXME: We should choose between various failure modes here
                #
                # Fail the bolometer by replacing the signal with
                # a linear trend.  We do *not* record the failure in the
                # detector flags so the data *cannot* be reduced without
                # some sort of mitigation
                start, stop = np.random.rand(2) * 10 * u.K
                x = (times - times[0]) / (times[-1] - times[0])
                trend = start + (stop - start) * x
                signal[det] = trend
                nfail += 1

        if comm is not None:
            nfail = comm.allreduce(nfail)
        ndet = len(all_dets)
        log.debug_rank(
            f"Failed {nfail} / {ndet} = {nfail/ndet:.3f} bolometers.  "
            f"Target yield was {yield_:.3f}",
            comm=comm,
        )
        return

    @function_timer
    def _misidentify_bolometers(
            self,
            comm,
            obs_id,
            times,
            focalplane,
            signal,
            all_dets,
            local_dets,
            det2rank,
    ):
        """Swap detector data between two confused readout channels"""
        if not self.misidentify_bolometers:
            return
        log = Logger.get()
        np.random.seed(int(obs_id + self.realization) % 2**32)
        nhit = 0
        for det1 in all_dets:
            for det2 in all_dets:
                if det1 == det2:
                    continue
                matched = True
                # Misidentification requires several properties to agree
                for prop in ["band", "card_slot", "wafer_slot", "bias", "pol"]:
                    if focalplane[det1][prop] != focalplane[det2][prop]:
                        matched = False
                        break
                if not matched:
                    continue
                # Detectors are matched, see if they get misidentified
                quat1 = focalplane[det1]["quat"]
                quat2 = focalplane[det2]["quat"]
                vec1 = qa.rotate(quat1, ZAXIS)
                vec2 = qa.rotate(quat2, ZAXIS)
                dist = np.arccos(np.dot(vec1, vec2))
                x = np.random.randn() * self.misidentification_width.to_value(u.rad)
                if np.abs(x) > dist:
                    nhit += 1
                    # Swap the signals.  This may require MPI communication
                    if det1 in local_dets and det2 in local_dets:
                        # Local operation
                        temp = signal[det1].copy()
                        signal[det1] = signal[det2]
                        signal[det2] = temp
                    elif det1 in local_dets:
                        # Send and receive signal
                        target = det2rank[det2]
                        comm.Send(signal[det1], dest=target, tag=nhit)
                        comm.Recv(signal[det1], source=target, tag=nhit)
                    elif det2 in local_dets:
                        # Receive and send signal
                        target = det2rank[det1]
                        sig2 = signal[det2].copy()
                        comm.Recv(signal[det2], source=target, tag=nhit)
                        comm.Send(sig2, dest=target, tag=nhit)
        ndet = len(all_dets)
        npair = ndet * (ndet - 1) // 2
        log.debug_rank(f"Misidentified {nhit} / {npair} detector pairs", comm=comm)
        return

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        for trait in "shared_flags", "det_data":
            value = getattr(self, trait)
            if value is None:
                raise RuntimeError(f"You must set `{trait}` before running SimReadout")

        for obs in data.obs:
            comm = obs.comm.comm_group
            obs_id = obs.uid
            times = obs.shared[self.times].data
            all_dets = obs.all_detectors
            local_dets = obs.select_local_detectors(detectors)
            obs.detdata.ensure(self.det_data, detectors=local_dets, create_units=u.K)

            # Map detectors to their owning processes
            det2rank = {}
            for rank, dets in enumerate(obs.dist.dets):
                for det in dets:
                    det2rank[det] = rank
            focalplane = obs.telescope.focalplane
            signal = obs.detdata[self.det_data]

            # Map bias lines to detectors
            bias2det = {}
            for det in local_dets:
                tube = focalplane[det]["tube_slot"]
                if tube not in bias2det:
                    bias2det[tube] = {}
                wafer = focalplane[det]["wafer_slot"]
                if wafer not in bias2det[tube]:
                    bias2det[tube][wafer] = {}
                bias = focalplane[det]["bias"]
                if bias not in bias2det[tube][wafer]:
                    bias2det[tube][wafer][bias] = []
                bias2det[tube][wafer][bias].append(det)

            # Map detectors to their physical position on the focalplane
            platescale = focalplane.detector_data.meta[
                "platescale"
            ].to_value(u.deg / u.mm)
            det2position = {}
            xrot = qa.rotation(YAXIS, np.pi / 2)
            for det in local_dets:
                quat = focalplane[det]["quat"]
                vec = qa.rotate(qa.mult(xrot, quat), ZAXIS)
                lon, lat = hp.vec2dir(vec, lonlat=True)
                x = lon / platescale
                y = lat / platescale
                det2position[det] = (x, y)

            self._add_glitches(
                comm,
                obs_id,
                times,
                focalplane,
                signal,
                all_dets,
                local_dets,
                bias2det,
                det2position,
            )
            self._add_jumps(
                comm, obs_id, times, focalplane, signal, all_dets, local_dets, bias2det
            )
            self._fail_bolometers(
                comm, obs_id, times, focalplane, signal, all_dets, local_dets, bias2det
            )
            self._misidentify_bolometers(
                comm, obs_id, times, focalplane, signal, all_dets, local_dets, det2rank
            )

        return

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "shared": [self.times, self.shared_flags],
            "detdata": [self.det_data],
        }
        return req

    def _provides(self):
        prov = {
            "meta": list(),
            "shared": [],
            "detdata": [self.det_data],
        }
        return prov

    def _accelerators(self):
        return list()



import numpy as np
import logging
logger = logging.getLogger(__name__)

def bin_signal(aman, bin_by, signal=None,
               range=None, bins=100, flags=None,
               weight_for_signal=None):
    """
    Bin time-ordered data by the ``bin_by`` and return the binned signal and its standard deviation.

    Parameters
    ----------
    aman : TOD
        The Axismanager object to be binned.
    bin_by : array-like
        The array by which signal is binned. It should has the same samps as aman.
    signal : str, optional
        The name of the signal to be binned. Defaults to aman.signal if not specified.
    range : list or None
        A list specifying the bin range ([min, max]). Default is None, which means bin range
        is set to [min(bin_by), max(bin_by)].
    bins : int or sequence of scalars
        If bins is an int, it defines the number of equal-width bins in the given range (100, by default).
        If bins is a sequence, it defines the bin edges, including the rightmost edge, allowing for non-uniform bin widths.
        If `bins` is a sequence, `bins` overwrite `range`.
    flags : (str or RangesMatrix or Ranges), optional
        Flag indicating whether to exclude flagged samples when binning the signal.
        If provided by a string, `aman.flags.get(flags)` is used for the flags.
        Default is no mask applied.
    weight_for_signal : array-like, optional
        Array of weights for the signal values. If None, all weights are assumed to be 1. You can get a apodizing window by
        'sotodlib.tod_ops.apodize.get_apodize_window_for_ends' or 'get_apodize_window_from_flags'.

    Returns
    -------
    Dictionary:
        - **bin_edges** (dict key): float array of bin edges length(bin_centers)+1.
        - **bin_centers** (dict key): center of each bin.
        - **bin_counts** (dict key): counts of binned samples. 
        - **binned_signal** (dict key): binned signal.
        - **binned_signal_sigma** (dict key): estimated sigma of binned signal.
    """
    if signal is None:
        signal = aman.signal
    if range is None:
        range = (np.nanmin(bin_by), np.nanmax(bin_by))

    signal_dtype = signal.dtype
    
    if weight_for_signal is None:
        weight_for_signal = np.ones(aman.samps.count, signal_dtype)
        
    # get bin_edges
    bin_edges = np.histogram_bin_edges(bin_by, bins=bins, range=range,)
    bin_centers = (bin_edges[1] - bin_edges[0])/2. + bin_edges[:-1] # edge to center
    nbins = len(bin_centers)
    
    # prepare binned signal array
    binned_signal = np.full([aman.dets.count, nbins], np.nan, signal_dtype)
    binned_signal_squared_mean = np.full([aman.dets.count, nbins], np.nan, signal_dtype)
    binned_signal_sigma = np.full([aman.dets.count, nbins], np.nan, signal_dtype)
    
    # get bin indices
    bin_indices = np.digitize(bin_by, bin_edges) - 1
    bin_indices = np.clip(bin_indices, 0, nbins-1)
        
    # bin tod
    if flags is None:
        bin_counts, _ = np.histogram(bin_by, bins=bins, range=range, weights = weight_for_signal)
        mcnts = bin_counts > 0
        
        for i, dets in enumerate(aman.dets.vals):      
            binned_signal[i][mcnts] = np.bincount(bin_indices, weights=signal[i]*weight_for_signal, minlength=nbins
                                                 )[mcnts]/bin_counts[mcnts]
            binned_signal_squared_mean[i][mcnts] = np.bincount(bin_indices, weights=(signal[i]*weight_for_signal)**2, minlength=nbins
                                                 )[mcnts]/bin_counts[mcnts]
            
        binned_signal_sigma[:, mcnts] = np.sqrt(np.abs(binned_signal_squared_mean[:,mcnts] - binned_signal[:,mcnts]**2)
                                      ) / np.sqrt(bin_counts[mcnts])
        bin_counts_dets = np.tile(bin_counts, (aman.dets.count, 1))

    else:
        bin_counts_dets = np.full([aman.dets.count, nbins], np.nan)
        if isinstance(flags, str):
            flags = aman.flags.get(flags)
        if flags.shape == (aman.dets.count, aman.samps.count):
            flag_is_2d = True
            m_2d = ~flags.mask()
        elif flags.shape == (aman.samps.count, ):
            flag_is_2d = False
            m = ~flags.mask()
        else:
            raise ValueError('flags should have shape of (`dets`, `samps`) or (`samps`,)')

        for i, dets in enumerate(aman.dets.vals):
            if flag_is_2d:
                m = m_2d[i]

            if weight_for_signal.shape == (aman.dets.count, aman.samps.count):
                weight_for_signal_det = weight_for_signal[i]
            elif weight_for_signal.shape == (aman.samps.count, ):
                weight_for_signal_det = weight_for_signal
            else:
                raise ValueError('weight_for_signal should have shape of (`dets`, `samps`) or (`samps`,)')

            bin_counts_dets[i] = np.bincount(bin_indices[m], weights=weight_for_signal_det[m], minlength=nbins)
            mcnts = bin_counts_dets[i] > 0
            binned_signal[i][mcnts] = np.bincount(bin_indices[m], weights=signal[i][m]*weight_for_signal_det[m], minlength=nbins
                                                 )[mcnts]/bin_counts_dets[i][mcnts]
            binned_signal_squared_mean[i][mcnts] = np.bincount(bin_indices[m], weights=(signal[i][m]*weight_for_signal_det[m])**2, minlength=nbins
                                                 )[mcnts]/bin_counts_dets[i][mcnts]
            binned_signal_sigma[i][mcnts] = np.sqrt(np.abs(binned_signal_squared_mean[i,mcnts] - binned_signal[i,mcnts]**2)
                                                 ) / np.sqrt(bin_counts_dets[i][mcnts])

    return {'bin_edges': bin_edges, 'bin_centers': bin_centers, 'bin_counts': bin_counts_dets,
            'binned_signal': binned_signal, 'binned_signal_sigma': binned_signal_sigma}



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Tools specific to ACT processing.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_act_responsivity_sign(operators):
    """Add commandline args and operators for ACT responsivity sign flip.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.ActSign(name="act_sign", enabled=False))


@workflow_timer
def act_responsivity_sign(job, otherargs, runargs, data):
    """Apply sign to ACT timestreams.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    job_ops = job.operators

    if job_ops.act_sign.enabled:
        job_ops.act_sign.apply(data)




import numpy as np
from collections import OrderedDict
import warnings


class ResultSet(object):
    """ResultSet is a special container for holding the results of
    database queries, i.e. columnar data.  The repr of a ResultSet
    states the name of its columns, and the number of rows::

      >>> print(rset)
      ResultSet<[array_code,freq_code], 17094 rows>

    You can access the column names in .keys::

      >>> print(rset.keys)
      ['array_code', 'freq_code']

    You can request a column by name, and a numpy array of values will
    be constructed for you:

      >>> rset['array_code']
      array(['LF1', 'LF1', 'LF1', ..., 'LF1', 'LF1', 'LF1'], dtype='<U3')

    You can request a row by number, and a dict will be constructed
    for you:

      >>> rset[10]
      {'base.array_code': 'LF1', 'base.freq_code': 'f027'}

    Note that the array or dict returned by indexing the ResultSet
    present copies of the data, not changing those objects will not
    update the original ResultSet.

    You can also access the raw row data in .rows, which is a simple
    list of tuples.  If you want to edit the data in a ResultSet,
    modify those data rows directly, or else use ``.asarray()`` to get
    a numpy array, modify the result, and create and a new ResultSet
    from that using the ``.from_friend`` constructor.

    You can get a structured numpy array using:

      >>> ret.asarray()
      array([('LF1', 'f027'), ('LF1', 'f027'), ('LF1', 'f027'), ...,
              ('LF1', 'f027'), ('LF1', 'f027'), ('LF1', 'f027')],
            dtype=[('array_code', '<U3'), ('freq_code', '<U4')])

    Slicing works along the row axis; and you can combine two results.
    So you could reorganize results like this, if you wanted:

      >>> rset[::2] + rset[1::2]
      ResultSet<[array_code,freq_code], 17094 rows>

    Finally, the .distinct() method returns a ResultSet containing the
    distinct elements:

      >>> rset.distinct()
      ResultSet<[array_code,freq_code], 14 rows>

    """

    #: Once instantiated, a list of the names of the ResultSet
    #: columns.
    keys = None

    #: Once instantiated, a list of the raw data tuples.
    rows = None

    def __init__(self, keys, src=None):
        self.keys = list(keys)
        if src is None:
            self.rows = []
        else:
            self.rows = [tuple(x) for x in src]

    @classmethod
    def from_friend(cls, source):
        """Return a new ResultSet populated with data from source.

        If source is a ResultSet, a copy is made.  If source is a
        numpy structured array, the ResultSet is constructed based on
        the dtype names and rows of source.

        Otherwise, a TypeError is raised.

        """
        if isinstance(source, np.ndarray):
            keys = source.dtype.names  # structured array?
            return cls(keys, list(source))
        if isinstance(source, ResultSet):
            return cls(source.keys, source.rows)
        raise TypeError(f"No implementation to construct {cls} from {source.__class__}.")

    def copy(self):
        return self.__class__(self.keys, self.rows)

    def subset(self, keys=None, rows=None):
        """Returns a copy of the object, selecting only the keys and rows
        specified.

        Arguments:
          keys: a list of keys to keep.  None keeps all.

          rows: a list or array of the integers representing which
            rows to keep.  This can also be specified as an array of
            bools, of the same length as self.rows, to select row by
            row.  None keeps all.

        """
        if keys is None:
            keys = self.keys
            def key_sel_func(row):
                return row
        else:
            key_idx = [self.keys.index(k) for k in keys]
            def key_sel_func(row):
                return [row[i] for i in key_idx]
        if rows is None:
            new_rows = map(key_sel_func, self.rows)
        elif isinstance(rows, np.ndarray) and rows.dtype == bool:
            assert(len(rows) == len(self.rows))
            new_rows = [key_sel_func(r) for r, s in zip(self.rows, rows) if s]
        else:
            new_rows = [key_sel_func(self.rows[i]) for i in rows]
        return self.__class__(keys, new_rows)

    @classmethod
    def from_cursor(cls, cursor, keys=None):
        """Create a ResultSet using the results stored in cursor, an
        sqlite.Cursor object.  The cursor must have be configured so
        that .description is populated.

        """
        if keys is None:
            keys = [c[0] for c in cursor.description]
        self = cls(keys)
        self.rows = [tuple(r) for r in cursor]
        return self

    def asarray(self, simplify_keys=False, hdf_compat=False):
        """Get a numpy structured array containing a copy of this data.  The
        names of the fields are taken from self.keys.

        Args:
          simplify_keys: If True, then the keys are stripped of any
            prefix (such as 'base.').  This is mostly for DetDb, where
            the table name can be annoying.  An error is thrown if
            this results in duplicate field names.
          hdf_compat: If True, then 'U'-type columns (Unicode strings)
            are converted to 'S'-type (byte strings), so it can be
            stored in an HDF5 dataset.

        """
        keys = [k for k in self.keys]
        if simplify_keys:  # remove prefixes
            keys = [k.split('.')[-1] for k in keys]
            assert(len(set(keys)) == len(keys))  # distinct.
        columns = tuple(map(_smart_array_cast, zip(*self.rows)))
        if hdf_compat:
            # Translate any Unicode columns to strings.
            new_cols = []
            for c in columns:
                if c.dtype.char == 'U':
                    new_cols.append(c.astype('S'))
                else:
                    new_cols.append(c)
            columns = new_cols
        dtype = [(k, c.dtype, c.shape[1:]) for k, c in zip(keys, columns)]
        output = np.ndarray(shape=len(columns[0]), dtype=dtype)
        for k, c in zip(keys, columns):
            output[k] = c
        return output

    def distinct(self):
        """
        Returns a ResultSet that is a copy of the present one, with
        duplicates removed.  The rows are sorted (according to python
        sort).
        """
        return self.__class__(self.keys, sorted(list(set(self.rows))))

    def strip(self, patterns=[]):
        """For any keys that start with a string in patterns, remove that
        string prefix from the key.  Operates in place.

        """
        for i, k in enumerate(self.keys):
            for p in patterns:
                if k.startswith(p):
                    self.keys[i] = k[len(p):]
                    break
        assert(len(self.keys) == len(set(self.keys)))

    def to_axismanager(self, axis_name="dets", axis_key="dets"):
        """Build an AxisManager directly from a ResultSet, projecting all columns
        along a single axis. This requires no additional metadata to build
        
        Args:
            axis_name: string, name of the axis in the AxisManager
            axis_key: string, name of the key in the ResultSet to put into the
                axis labels. This key will not be added to the AxisManager
                fields. 
        """
        from sotodlib import core
        aman = core.AxisManager(
            core.LabelAxis(axis_name, self[axis_key])
        )
        for k in self.keys:
            if k == axis_key:
                continue
            if any([ x is None for x in self[k]]):
                raise TypeError("None(s) found in key {}, these cannot be ".format(k)+
                               "nicely wrapped into an AxisManager")
            aman.wrap(k, self[k], [(0,axis_name)])
        return aman

    def restrict_dets(self, restriction, detdb=None):
        # There are 4 classes of keys:
        # - dets:* keys appearing only in restriction
        # - dets:* keys appearing only in self
        # - dets:* keys appearing in both
        # - other.
        new_keys = [k for k in restriction if k.startswith('dets:')]
        match_keys = []
        for k in self.keys:
            if k in new_keys:
                match_keys.append(k)
                new_keys.remove(k)
        other_keys = [k for k in self.keys if k not in match_keys]
        output_keys = new_keys + match_keys + other_keys # disjoint.
        output_rows = []
        for row in self:
            row = dict(row)  # copy
            for k in match_keys:
                if row[k] != restriction[k]:
                    break
            else:
                # You passed.
                row.update({k: restriction[k] for k in new_keys})
            output_rows.append([row[k] for k in output_keys])
        # That's all.
        return self.__class__(output_keys, output_rows)

    # Everything else is just implementing container-like behavior

    def __repr__(self):
        keystr = 'empty'
        if self.keys is not None:
            keystr = ','.join(self.keys)
        return ('{}<[{}], {} rows>'.format(self.__class__.__name__,
                                           keystr, len(self)))

    def __len__(self):
        return len(self.rows)

    def append(self, item):
        vals = []
        for k in self.keys:
            if k not in item.keys():
                raise ValueError(f"Item to append must include key '{k}'")
            vals.append(item[k])
        self.rows.append(tuple(vals))

    def extend(self, items):
        if not isinstance(items, ResultSet):
            raise TypeError("Extension only valid for two ResultSet objects.")
        if self.keys != items.keys:
            raise ValueError("Keys do not match: {} <- {}".format(
                self.keys, items.keys))
        self.rows.extend(items.rows)

    def __getitem__(self, item):
        # Simple row look-up... convert to dict.
        if isinstance(item, int) or isinstance(item, np.integer):
            return OrderedDict([(k,v) for k, v in
                                zip(self.keys, self.rows[item])])
        # Look-up by column...
        if isinstance(item, str):
            index = self.keys.index(item)
            return _smart_array_cast([x[index] for x in self.rows],
                                     field_detail=f"Key {item}:")
        # Slicing.
        output = self.__class__(self.keys, self.rows[item])
        return output

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __add__(self, other):
        output = self.copy()
        output += other
        return output

    @staticmethod
    def concatenate(items, axis=0):
        assert(axis == 0)
        output = items[0].copy()
        for item in items[1:]:
            output += item
        return output

    def merge(self, src):
        """Merge with src, which must have same number of rows as self.
        Duplicate columns are not allowed.

        """
        if len(self) != len(src):
            raise ValueError("self and src have different numbers of rows.")
        for k in src.keys:
            if k in self.keys:
                raise ValueError("Duplicate key: %s" % k)
        new_keys = self.keys + src.keys
        new_rows = [r0 + r1 for r0, r1 in zip(self.rows, src.rows)]
        self.keys, self.rows = new_keys, new_rows


def _smart_array_cast(values, dtype=None, field_detail=None):
    """Convert a list of values to a numpy array.  Let numpy casting do
    its job, but replace any Nones in the list with some better value,
    first.  The better value is nan, for floats.  For strings and
    ints, the best we can do is '' and 0 -- in which case a warning is
    issued.

    For string and int table columns it would be much better to set
    sensible default values in the schema rather than have null vals
    persist in the db.

    """
    non_null = [v for v in values if v is not None]
    if len(non_null) == 0:
        return np.full(len(values), np.nan)
    if len(non_null) == len(values):
        return np.array(values, dtype=dtype)

    warn = None
    trial = np.array(non_null)
    if trial.dtype == np.float32:
        fill = np.float32(np.nan)
    elif trial.dtype == np.float64:
        fill = np.float64(np.nan)
    elif np.issubdtype(trial.dtype, np.str_):
        fill = ''
        warn = 'Replacing null entries with "".'
    elif np.issubdtype(trial.dtype, np.integer):
        fill = 0
        warn = 'Replacing null entries with 0.'
    else:
        fill = None
        warn = 'No patch value for null (dtype=%s).' % (str(trial.dtype))

    revalues = np.array([fill if v is None else v
                         for v in values])
    if revalues.dtype != trial.dtype:
        warn = ('' if warn is None else warn + ' ') + \
        warnings.warn("Unexpected dtype change.")

    if warn:
        if field_detail:
            warn = field_detail + ' ' + warn
        warnings.warn(warn)

    return revalues



#!/usr/bin/env python3

# Copyright (c) 2019-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""
This workflow runs basic data reduction and map-making:

- A single MPI process group is used

- All input observations are loaded and made into the output map

In particular, this script is designed for testing mapmaking techniques on data that
already exists on disk.  It does not simulate data and it is not a full workflow for
running null tests, building observation matrices, etc.

You can see the automatically generated command line options with:

    toast_so_map --help

Or you can dump a config file with all the default values with:

    toast_so_map --default_toml config.toml

This script contains just comments about what is going on.  For details about all the
options for a specific Operator, see the documentation or use the help() function from
an interactive python session.

"""

import argparse
import datetime
import os
import sys
import traceback

import numpy as np

from astropy import units as u

# Import sotodlib.toast first, since that sets default object names
# to use in toast.
import sotodlib.toast as sotoast

import toast
import toast.ops
from toast.mpi import MPI, Comm
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .. import workflows as wrk

# Make sure pixell uses a reliable FFT engine
import pixell.fft

pixell.fft.engine = "fftw"


def reduce_data(job, otherargs, runargs, data):
    log = toast.utils.Logger.get()
    job_ops = job.operators
    job_tmpl = job.templates

    # Preprocess data to get flags / cuts
    wrk.preprocess(job, otherargs, runargs, data)

    # Now apply preprocessing to good detectors
    wrk.raw_statistics(job, otherargs, runargs, data)

    wrk.diff_noise_estimation(job, otherargs, runargs, data)
    wrk.noise_estimation(job, otherargs, runargs, data)

    data = wrk.demodulate(job, otherargs, runargs, data)

    wrk.filter_ground(job, otherargs, runargs, data)
    wrk.filter_common_mode(job, otherargs, runargs, data)
    wrk.filter_poly1d(job, otherargs, runargs, data)
    wrk.filter_poly2d(job, otherargs, runargs, data)

    wrk.diff_noise_estimation(job, otherargs, runargs, data)
    wrk.noise_estimation(job, otherargs, runargs, data)

    wrk.processing_mask(job, otherargs, runargs, data)
    wrk.flag_sso(job, otherargs, runargs, data)
    wrk.hn_map(job, otherargs, runargs, data)
    wrk.cadence_map(job, otherargs, runargs, data)
    wrk.crosslinking_map(job, otherargs, runargs, data)

    if job.operators.splits.enabled:
        wrk.splits(job, otherargs, runargs, data)
    else:
        wrk.mapmaker_ml(job, otherargs, runargs, data)
        wrk.mapmaker(job, otherargs, runargs, data)
        wrk.mapmaker_filterbin(job, otherargs, runargs, data)
        wrk.mapmaker_madam(job, otherargs, runargs, data)
    wrk.filtered_statistics(job, otherargs, runargs, data)

    mem = toast.utils.memreport(
        msg="(whole node)", comm=data.comm.comm_world, silent=True
    )
    log.info_rank(f"After reducing data:  {mem}", data.comm.comm_world)


def main():
    env = toast.utils.Environment.get()
    log = toast.utils.Logger.get()
    gt = toast.timing.GlobalTimers.get()
    gt.start("toast_so_map (total)")
    timer = toast.timing.Timer()
    timer.start()

    # Get optional MPI parameters
    comm, procs, rank = toast.get_world()

    if "OMP_NUM_THREADS" in os.environ:
        nthread = os.environ["OMP_NUM_THREADS"]
    else:
        nthread = 1
    log.info_rank(
        f"Executing workflow with {procs} MPI tasks, each with "
        f"{nthread} OpenMP threads at {datetime.datetime.now()}",
        comm,
    )

    mem = toast.utils.memreport(msg="(whole node)", comm=comm, silent=True)
    log.info_rank(f"Start of the workflow:  {mem}", comm)

    parser = argparse.ArgumentParser(description="SO mapmaking pipeline")

    parser.add_argument(
        "--out_dir",
        required=False,
        type=str,
        default="toast_out",
        help="The output directory",
    )
    parser.add_argument(
        "--obsmaps",
        required=False,
        default=False,
        action="store_true",
        help="Map each observation separately.",
    )
    parser.add_argument(
        "--detmaps",
        required=False,
        default=False,
        action="store_true",
        help="Map each detector separately.",
    )
    parser.add_argument(
        "--intervalmaps",
        required=False,
        default=False,
        action="store_true",
        help="Map each interval separately.",
    )
    parser.add_argument(
        "--preprocess_copy",
        required=False,
        default=False,
        action="store_true",
        help="Perform preprocessing on a copy of the data.",
    )
    parser.add_argument(
        "--log_config",
        required=False,
        default=None,
        help="Dump out config log to this yaml file",
    )

    # The operators and templates we want to configure from the command line
    # or a parameter file.

    operators = list()
    templates = list()

    wrk.setup_load_or_simulate_observing(parser, operators)
    wrk.setup_preprocess(parser, operators)

    wrk.setup_raw_statistics(operators)

    wrk.setup_demodulate(operators)

    wrk.setup_filter_ground(operators)
    wrk.setup_filter_common_mode(operators)
    wrk.setup_filter_poly1d(operators)
    wrk.setup_filter_poly2d(operators)

    wrk.setup_noise_estimation(operators)

    wrk.setup_processing_mask(operators)
    wrk.setup_flag_sso(operators)
    wrk.setup_hn_map(operators)
    wrk.setup_cadence_map(operators)
    wrk.setup_crosslinking_map(operators)

    wrk.setup_mapmaker_ml(operators)
    wrk.setup_mapmaker(operators, templates)
    wrk.setup_mapmaker_filterbin(operators)
    wrk.setup_mapmaker_madam(operators)
    wrk.setup_splits(operators)
    wrk.setup_filtered_statistics(operators)

    job, config, otherargs, runargs = wrk.setup_job(
        parser=parser, operators=operators, templates=templates
    )

    # Create our output directory
    if comm is None or comm.rank == 0:
        if not os.path.isdir(otherargs.out_dir):
            os.makedirs(otherargs.out_dir, exist_ok=True)

    # Log the config that was actually used at runtime.
    if otherargs.log_config is not None:
        toast.config.dump_yaml(otherargs.log_config, config, comm=comm)

    # If this is a dry run, exit
    if otherargs.dry_run:
        log.info_rank("Dry-run complete", comm=comm)
        return

    # Load data
    data = wrk.load_or_simulate_observing(job, otherargs, runargs, comm)

    # Reduce it
    reduce_data(job, otherargs, runargs, data)

    # Collect optional timing information
    alltimers = toast.timing.gather_timers(comm=comm)
    if data.comm.world_rank == 0:
        out = os.path.join(otherargs.out_dir, "timing")
        toast.timing.dump(alltimers, out)

    log.info_rank("Workflow completed in", comm=comm, timer=timer)


def cli():
    world, procs, rank = toast.mpi.get_world()
    with toast.mpi.exception_guard(comm=world):
        main()


if __name__ == "__main__":
    cli()



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Test toast data export.
"""

import os
import shutil
import glob
import re

import numpy as np
import astropy.units as u

from unittest import TestCase

# Import so3g before any other packages that import spt3g
import so3g

from ._helpers import create_outdir, simulation_test_multitube, close_data_and_comm


try:
    # Import sotodlib toast module first, which sets global toast defaults
    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    import toast
    from toast.observation import default_values as defaults

    toast_available = True
except ImportError:
    raise
    toast_available = False


class ToastBooksTest(TestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        if not toast_available:
            print("TOAST cannot be imported- skipping unit tests", flush=True)
            return
        world, procs, rank = toast.get_world()
        self.outdir = create_outdir(fixture_name, mpicomm=world)

    def test_book_saveload(self):
        if not toast_available:
            return
        world, procs, rank = toast.get_world()
        testdir = os.path.join(self.outdir, "save_load")
        if world is None or world.rank == 0:
            if os.path.isdir(testdir):
                shutil.rmtree(testdir)
            os.makedirs(testdir)
        if world is not None:
            world.barrier()

        data = simulation_test_multitube(
            world,
            telescope_name="LAT",
            tubes="c1,i6",
            sample_rate=2.0 * u.Hz,
            detset_key="pixel",
            thin_fp=100,
            temp_dir=self.outdir,
        )

        # Create a noise model from focalplane detector properties
        noise_model = toast.ops.DefaultNoiseModel()
        noise_model.apply(data)

        # Simulate some noise
        sim_noise = toast.ops.SimNoise()
        sim_noise.apply(data)

        # Corotation
        corotator = so_ops.CoRotator()
        corotator.apply(data)

        # Purge intervals.  The book format currently does not support things
        # like this.  We would have to reconstruct those from the flag information
        # for now.  Make a copy for later comparison

        original = list()
        for ob in data.obs:
            interval_names = list(ob.intervals.keys())
            for iname in interval_names:
                del ob.intervals[iname]
            original.append(ob.duplicate(times="times"))

        # Set up the save operator

        book_dir = os.path.join(testdir, "books")
        focalplane_dir = os.path.join(testdir, "focalplanes")
        noise_dir = os.path.join(testdir, "noise_models")

        save = so_ops.SaveBooks(
            book_dir=book_dir,
            focalplane_dir=focalplane_dir,
            noise_dir=noise_dir,
            frame_intervals=None,
            gzip=True,
            hwp_angle=None, # This is LAT data
        )

        # Save the data
        # data.info()
        save.apply(data)

        # Load the data

        new_data = toast.Data(comm=data.comm)
        load = so_ops.LoadBooks(
            books=glob.glob(os.path.join(book_dir, "obs_*")),
            focalplane_dir=focalplane_dir,
            noise_dir=noise_dir,
            detset_key="pixel",
            hwp_angle=None,
            corotator_angle=defaults.corotator_angle,
            boresight_angle=None,
        )
        load.apply(new_data)
        # new_data.info()

        # Check for consistency

        # print(f"Original names:  {[x.name for x in data.obs]}", flush=True)

        # print(f"New names:  {[x.name for x in new_data.obs]}", flush=True)

        new_order = {y.name: x for x, y in enumerate(new_data.obs)}

        for orig in original:
            # The telescope name in the original data is modified by the observing
            # split.  It does not matter in practice, but causes the equality
            # test to fail.
            orig.telescope.name = re.sub(r"^LAT_", "", orig.telescope.name)

            ob = new_data.obs[new_order[orig.name]]

            # FIXME:  Leave these in if we want to check general metadata
            # handling.
            del orig["scan_el"]
            del orig["scan_min_az"]
            del orig["scan_max_az"]
            del orig["scan_min_el"]
            del orig["scan_max_el"]

            # The loaded noise model is a base class instance, and the one
            # written is an analytic model.  Fix this eventually so that
            # the comparison passes below.
            del orig["noise_model"]
            del ob["noise_model"]

            if ob != orig:
                print(f"-------- Proc {data.comm.world_rank} ---------\n{orig}\n{ob}")
                self.assertTrue(False)

        close_data_and_comm(new_data)
        close_data_and_comm(data)



import copy
import numpy as np
import logging
from scipy.special import eval_legendre
from sotodlib.tod_ops import flags
logger = logging.getLogger(__name__)

def subscan_polyfilter(aman, degree, signal_name="signal", exclude_turnarounds=False, 
                       mask=None, exclusive=True, method="legendre", in_place=True):
    """
    Apply polynomial filtering to subscan segments in a data array.
    This function applies polynomial filtering to subscan segments within signal for each detector.
    Subscan segments are defined based on the presence of flags such as 'left_scan' and 'right_scan'. Polynomial filtering
    is used to remove low-degree polynomial trends within each subscan segment.

    Arguments
    ---------
    aman : AxisManager
    degree : int
        The degree of the polynomial to be removed.
    signal_name : string, optional
        The name of TOD signal to use. If not provided, `aman.signal` will be used.
    exclude_turnarounds : bool
        Optional. If True, turnarounds are excluded from subscan identification. Default is False.
    mask : str or RangesMatrix
        Optional. A mask used to select specific data points for filtering. 
        If None, no mask is applied. If the mask is given in str, ``aman.flags['mask']`` is used as mask.
        Arbitrary mask can be specified in the style of RangesMatrix.
    exclusive :
        Optional. If True, the mask is used to exclude data from fitting. If False, the mask is used to include data for fitting.
        Default is True.
    method : str
        Optioal. Method to model the baseline of TOD.
        In `legendre` method, baseline model is constructed using orthonormality of Legendre function.
        In `polyfit` method, numpy.polyfit is used.
        `legendre` is faster. Default is `legendre`.
    in_place: bool
        Optional. If True, `aman.signal` is overwritten with the processed signal.

    Returns
    -------
    signal : array-like
        The processed signal.
    """
    if method not in ["polyfit", "legendre"] :
        raise ValueError("Only polyfit and legendre are acceptable.")
        
    if exclude_turnarounds:
        if ("left_scan" not in aman.flags) or ("turnarounds" not in aman.flags):
            logger.warning('aman does not have left/right scan or turnarounds flag. `sotodlib.flags.get_turnaround_flags` will be ran with default parameters')
            _ = flags.get_turnaround_flags(aman,truncate=True)
        ls_mask = aman.flags["left_scan"].mask()
        rs_mask = aman.flags["right_scan"].mask()
        ta_mask = aman.flags["turnarounds"].mask()
        
        valid_scan = np.logical_and(np.logical_or(ls_mask if ls_mask.ndim==1 else ls_mask[0], 
                                                  rs_mask if rs_mask.ndim==1 else rs_mask[0], ),
                                    ~ta_mask if ta_mask.ndim==1 else ta_mask[0])
        subscan_indices = _get_subscan_range_index(valid_scan)
    else:
        if ("left_scan" not in aman.flags):
            logger.warning('aman does not have left/right scan. `sotodlib.flags.get_turnaround_flags` will be ran with default parameters')
            _ = flags.get_turnaround_flags(aman,truncate=True)
            
        ls_mask = aman.flags["left_scan"].mask()
        rs_mask = aman.flags["right_scan"].mask()
        ta_mask = aman.flags["turnarounds"].mask()
        
        subscan_indices_l = _get_subscan_range_index(ls_mask if ls_mask.ndim==1 else ls_mask[0])
        subscan_indices_r = _get_subscan_range_index(rs_mask if rs_mask.ndim==1 else rs_mask[0])
        subscan_indices = np.vstack([subscan_indices_l, subscan_indices_r])
        subscan_indices= subscan_indices[np.argsort(subscan_indices[:, 0])]
        
    if subscan_indices[0][0] != 0:
        subscan_indices = np.vstack([ [0, subscan_indices[0][0]], subscan_indices])
    if subscan_indices[-1][-1] != aman.samps.count:
        subscan_indices = np.vstack([ subscan_indices, [subscan_indices[-1][-1], aman.samps.count]])
    
    signal = aman[signal_name]
    if not(in_place):
        signal = signal.copy()

    if mask is None:
        mask_array = np.zeros(aman.samps.count, dtype=bool)
    elif type(mask) is str:
        mask_array = aman.flags[mask].mask()
    else:
        mask_array = mask.mask()
    if exclusive is False:
        mask_array = ~mask_array
        
    is_matrix = len(mask_array.shape) > 1
    if method == "polyfit":
        t = aman.timestamps - aman.timestamps[0]
        for i_det in range(aman.dets.count):
            if is_matrix:
                each_det_mask = mask_array[i_det]
            else:
                each_det_mask = mask_array

            for start, end in subscan_indices:
                if np.count_nonzero(~each_det_mask[start:end]) < degree:
                    # If degree of freedom is lower than zero, just subtract mean
                    logger.warning('polyfit degree is smaller than the number of valid data points')
                    signal[i_det, start:end] -= np.mean(signal[i_det, start:end])
                else:
                    t_mean = np.mean(t[start:end])
                    pars = np.ma.polyfit(
                            np.ma.array(t[start:end]-t_mean, mask=each_det_mask[start:end]),
                            np.ma.array(signal[i_det,start:end], mask=each_det_mask[start:end]),
                            deg=degree)

                    signal[i_det,start:end] -= np.polyval(pars, t[start:end]-t_mean)

    elif method == "legendre":
        degree_corr = degree + 1

        time = np.copy(aman["timestamps"])

        for start, end in subscan_indices:
            ### Normalization constant of legendre function 
            norm_vector = np.arange(degree_corr)
            norm_vector = 2./(2.*norm_vector+1)
            
            # Get each subscan to be filtered
            tod_mat = copy.deepcopy(signal[:, start:end])


            # Scale time range into [-1,1]
            x = np.linspace(-1, 1, tod_mat.shape[1])
            dx = np.mean(np.diff(x))
            sub_time = time[start:end]

            # Generate legendre functions of each degree and store them in an array
            arr_legendre = np.array([eval_legendre(deg, x) for deg in range(degree_corr)])

            # flag to know if the result is matrix formation
            flag_matrix = False

            # Modify TODs if mask is defined
            if mask is None :
                pass
            else :
                # if mask is matrix like, we should interpolate TOD det by det.
                if is_matrix :
                    if np.sum((mask_array[:,start:end]).astype(np.int32)) > 0 :
                        msk_indx = mask_array[:,start:end]                        
                        for idet in range(tod_mat.shape[0]) : 
                            n_intep =  np.sum((mask_array[idet,start:end]).astype(np.int32))
                            if n_intep > 0:
                                if n_intep == tod_mat.shape[1] : continue
                                interped = np.interp(np.flatnonzero(msk_indx[idet]),np.flatnonzero(~msk_indx[idet]), tod_mat[idet][~msk_indx[idet]])
                                tod_mat[idet,msk_indx[idet]] = interped                            
                    else:
                        # If mask does not affect this range, just go through.
                        pass
                    
                # If mask is array like, same ranges of each det will be interpolated.
                else :
                    n_intep =  np.sum((mask_array[start:end]).astype(np.int32))
                    if n_intep > 0 :
                        if n_intep == tod_mat.shape[1] : continue
                        msk_indx = mask_array[start:end]
                        for idet in range(tod_mat.shape[0]) : 
                            interped = np.interp(np.flatnonzero(msk_indx),np.flatnonzero(~msk_indx), tod_mat[idet][~msk_indx])
                            tod_mat[idet,msk_indx] = interped
                    else :
                        pass
            
            
            means = np.mean(tod_mat, axis=1)[:, np.newaxis]
            tod_mat -= means
            
            # Make model to be subtracted
            coeffs = np.dot(arr_legendre, tod_mat.T)
            model = np.dot((coeffs/norm_vector[:, np.newaxis]).T,arr_legendre)*dx

            model += means
            signal[:,start:end] -= model

    if in_place:
        aman[signal_name] = signal

    return signal
                
def _get_subscan_range_index(scan_flag, _min=0):
    """
    Get the indices of subscans in a binary flag array.
    This function identifies subscan ranges within a binary flag array, where subscans are defined as consecutive
    sequences of 'True' values in the input 'scan_flag' array.

    Parameters:
    - scan_flag (numpy.ndarray): A 1-dimensional binary flag array indicating the presence of subscans.
    - _min (int, optional): The minimum length of a subscan range to consider. Default is 0.

    Returns:
    - numpy.ndarray: A 2-column array containing the start and end indices of subscan ranges in 'scan_flag' where
      the length of each subscan is greater than or equal to '_min'.
    """
    ones = np.where(scan_flag)[0]
    diff = np.diff(ones)
    
    starts = np.insert(ones[np.where(diff != 1)[0]+1], 0, ones[0])
    ends = np.append(ones[np.where(diff != 1)[0] ], ones[-1])
    indices = list(zip(starts, ends))
    indices = np.array([(start, end+1) for start, end in indices if (end - start + 1 >= _min)])
    return indices
 
